#[allow(clippy::module_inception, dead_code)]
pub mod bot {

    use std::result;

    use log::{info, debug};
    use matrix_sdk::{
        Client,
        config::{SyncSettings, RequestConfig},
        event_handler::Ctx,
        room::{Joined, Room, RoomMember},
        ruma::{
            UserId, ServerName, RoomId, OwnedEventId,
            events::room::message::{OriginalSyncRoomMessageEvent, RoomMessageEventContent}, 
        }
    };

    use sled;
    use chrono::{DateTime, Utc};
    use crate::utils::utils::*;
    use crate::config_reader::config_reader::BotUserInfo;

    #[allow(dead_code)]
    #[derive(Clone)]
    pub struct Bot {
        client: Client,
        info: BotUserInfo,
        joined_room: Joined,
        swear_list: Vec<String>,
        database_handle: sled::Db,
        spam_info: SpamInfo
    }

    impl Bot {
        pub async fn bot_login(creds: BotUserInfo) -> Self {
            let user_id = creds.user_id.as_str();
            let password = creds.password.as_str();
            let room = RoomId::parse(creds.room.as_str()).unwrap();
            let user = UserId::parse(user_id).expect("Unable to parse bot user id");
            let server_name = ServerName::parse(user.server_name()).unwrap();
            let request_config = RequestConfig::new().force_auth();
            let client = Client::builder()
                .server_name(&server_name)
                .request_config(request_config)
                .build()
                .await.
                unwrap();

            info!("Logging in…");
            let response = client
                .login_username(user.localpart(), password)
                .initial_device_display_name("Matrix-Modbot")
                .send()
                .await
                .expect("Unable to login");

            info!("Doing the initial sync…");
            client
                .sync_once(SyncSettings::new())
                .await
                .expect("Unable to sync");

            info!(
                "Logged in as {}, got device_id {}",
                response.user_id, response.device_id
            );
            
            // Code snippet from hebbot https://github.com/haecker-felix/hebbot
            // Try to accept room invite, if any
            if let Some(invited_room) = client.get_invited_room(&room) {
                invited_room
                    .accept_invitation()
                    .await
                    .expect("Matrix-Modbot could not join the invited room");
            }

            // Need to do another sync to make sure bot knows the newly joined rooms
            client
                .sync_once(SyncSettings::new())
                .await
                .expect("Unable to sync");

            let joined_room = client
                .get_joined_room(&room)
                .expect("Cannot get joined room");

            let swear_list = create_swear_list(creds.swear_list_url.as_str())
                .expect("Could not download swear list");

            // Syncs members list of the joined room
            joined_room.sync_members().await.unwrap();

            let members_list = client
                .get_joined_room(joined_room.room_id())
                .unwrap()
                .joined_members()
                .await
                .expect("Cannot get members");

            let path = "members_database";
            let db = sled::open(path).unwrap();
            let default_reputation: i64 = 0;
            let utc: DateTime<Utc> = Utc::now();
            let timestamp: i64 = utc.timestamp();
            let bytes = convert_to_bytes_sled(timestamp, default_reputation);
            for member in members_list {
                dbg!(
                    db.insert(member.user_id().as_str(), &bytes).unwrap()
                );
            };

            let bot = Bot {
                client,
                info: creds,
                joined_room,
                swear_list,
                database_handle: db,
                spam_info: SpamInfo::new()
            };

            bot.send_message("Sucessfully started Matrix-Modbot").await;
            
            bot.client.add_event_handler_context(bot.clone());
            bot.client.add_event_handler(Self::on_room_message);

            info!("Started syncing");
            bot.client.sync(SyncSettings::new()).await.unwrap();

            bot
        }

        async fn on_room_message(event: OriginalSyncRoomMessageEvent, room: Room, Ctx(bot): Ctx<Bot>) {
            if let Room::Joined(_joined) = &room {
                // Standard text message
                if let Some(text) = get_message_event_text(&event) {
                    let member = room.get_member(&event.sender).await.unwrap().unwrap();
                    let id = &event.event_id;

                    if bot.info.allow_swear && detect_swear_from_message(&bot.swear_list, &text) {
                            bot.delete_message_from_room(id, "Swearing").await;
                            bot.update_reputation_for_member(&member, -1).await.expect("Failed to subtract reputation from member");
                        };
                    }
                }

                // Message edit
                if let Some((edited_msg_event_id, text)) = get_edited_message_event_text(&event){
                    if bot.info.allow_swear && detect_swear_from_message(&bot.swear_list, &text) {
                        let member = room.get_member(&event.sender).await.unwrap().unwrap();
                        bot.delete_message_from_room(&edited_msg_event_id, "Swearing").await;
                        bot.update_reputation_for_member(&member, -1).await.expect("Failed to subtract reputation from member");
                    };
                }
        }

        async fn delete_message_from_room(&self, event_id: &OwnedEventId, reason: &str) {
            let result = self.joined_room.redact(event_id, Some(reason), None).await;
            if result.is_err() {
                self.send_message("Error deleting message").await;
            };
            self.send_message(&format!("{} is not permitted on this server", reason)).await;
        }

        async fn update_reputation_for_member(&self, member: &RoomMember, rep_change: i64) -> sled::Result<()> {
            let member_id: &str = member.user_id().as_str();
            let user_data = dbg!(self.database_handle.get(member_id)?);
            let (timestamp, mut reputation) = convert_from_bytes_sled(&user_data.unwrap());
            if reputation == -15 {
                self.kick_user(member, "Reputation too low!").await;
                Ok(())
            }
            else {
                reputation += rep_change;
                let new_data = convert_to_bytes_sled(timestamp, reputation);
                dbg!(self.database_handle.remove(member_id)?);
                dbg!(self.database_handle.insert(member_id, &new_data)?);
                self.send_message(&format!("{}'s current reputation is {}", member_id, reputation)).await;
                Ok(())
            }
        }

        async fn kick_user(&self, member: &RoomMember, reason: &str) {
            let member_id: &str = member.user_id().as_str();
             if member.power_level() <= 50 { // Won't kick mods and admins
                if let Some(_) = self.joined_room.kick_user(member.user_id(), Some(reason)).await.ok() {
                    dbg!(self.database_handle.remove(member_id).unwrap());
                    self.send_message(&format!("Member {} has been kicked.", member_id)).await;
                };
            }
            else {
                self.send_message("Cannot kick moderators and admins").await;
            }
        }
            
        async fn send_message(&self, message: &str) {
            debug!("Send message ({:?}): {}", &self.joined_room, message);
            let content = RoomMessageEventContent::text_plain(message);
            let _ = &self.joined_room.send(content, None).await.unwrap();
        }

        async fn detect_spam(&mut self, event: &OriginalSyncRoomMessageEvent) {
            let author = self.joined_room.get_member(&event.sender).await.unwrap().unwrap();
            let author_name = author.user_id().as_str().to_string();
            let curr_utc = Utc::now().timestamp();
            if self.spam_info.author_msg_times.contains_key(&author_name) {
                self.spam_info.author_msg_times.insert(author_name.clone(), vec![]);
            }
            else {
                self.spam_info.author_msg_times.get_mut(&author_name).unwrap().push(curr_utc);
            }

            let expire_time: i64 = curr_utc - self.spam_info.detection_window;

            let mut expired_msgs: Vec<i64> = vec![];
            for msg_time in self.spam_info.author_msg_times.get(&author_name).unwrap() {
                if msg_time < &expire_time {
                    expired_msgs.push(*msg_time)
                }
            }

            for msg in expired_msgs {
                self.spam_info.author_msg_times.get_mut(&author_name).unwrap().retain(|value| *value != msg);
            }

            if i64::try_from(self.spam_info.author_msg_times.get(&author_name).unwrap().len()).unwrap() > self.spam_info.max_msg_per_window {
                self.delete_message_from_room(&event.event_id, "Spamming").await;
            }
            
        }

        async fn detect_command(&self, event_id: &OwnedEventId) {
            todo!()
        }

    }
}
